@@ -1,9 +1,9 @@
 /*
- * "$Id: mxml-file.c 210 2005-05-19 14:56:05Z mike $"
+ * "$Id: mxml-file.c 312 2007-10-03 06:25:07Z mike $"
  *
  * File loading code for Mini-XML, a small XML-like file parsing library.
  *
- * Copyright 2003-2005 by Michael Sweet.
+ * Copyright 2003-2007 by Michael Sweet.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -24,8 +24,15 @@
  *   mxmlSaveFd()            - Save an XML tree to a file descriptor.
  *   mxmlSaveFile()          - Save an XML tree to a file.
  *   mxmlSaveString()        - Save an XML node tree to a string.
+ *   mxmlSAXLoadFd()         - Load a file descriptor into an XML node tree
+ *                             using a SAX callback.
+ *   mxmlSAXLoadFile()       - Load a file into an XML node tree
+ *                             using a SAX callback.
+ *   mxmlSAXLoadString()     - Load a string into an XML node tree
+ *                             using a SAX callback.
  *   mxmlSetCustomHandlers() - Set the handling functions for custom data.
  *   mxmlSetErrorCallback()  - Set the error message callback.
+ *   mxmlSetWrapMargin()     - Set the the wrap margin when saving XML data.
  *   mxml_add_char()         - Add a character to a buffer, expanding as needed.
  *   mxml_fd_getc()          - Read a character from a file descriptor.
  *   mxml_fd_putc()          - Write a character to a file descriptor.
@@ -48,8 +55,7 @@
  * Include necessary headers...
  */
 
-#include "config.h"
-#include "mxml.h"
+#include "mxml-private.h"
 #ifdef WIN32
 #  include <io.h>
 #else
@@ -74,31 +80,19 @@
 
 
 /*
- * Structures...
+ * Types and structures...
  */
 
-typedef struct mxml_fdbuf_s		/**** File descriptor buffer (@private) ****/
+typedef int (*_mxml_getc_cb_t)(void *, int *);
+typedef int (*_mxml_putc_cb_t)(int, void *);
+
+typedef struct _mxml_fdbuf_s		/**** File descriptor buffer ****/
 {
   int		fd;			/* File descriptor */
   unsigned char	*current,		/* Current position in buffer */
 		*end,			/* End of buffer */
 		buffer[8192];		/* Character buffer */
-} mxml_fdbuf_t;
-
-
-/*
- * Global error handler...
- */
-
-extern void	(*mxml_error_cb)(const char *);
-
-
-/*
- * Custom data handlers...
- */
-
-static mxml_custom_load_cb_t	mxml_custom_load_cb = NULL;
-static mxml_custom_save_cb_t	mxml_custom_save_cb = NULL;
+} _mxml_fdbuf_t;
 
 
 /*
@@ -109,32 +103,38 @@ static int		mxml_add_char(int ch, char **ptr, char **buffer,
 			              int *bufsize);
 static int		mxml_fd_getc(void *p, int *encoding);
 static int		mxml_fd_putc(int ch, void *p);
-static int		mxml_fd_read(mxml_fdbuf_t *buf);
-static int		mxml_fd_write(mxml_fdbuf_t *buf);
+static int		mxml_fd_read(_mxml_fdbuf_t *buf);
+static int		mxml_fd_write(_mxml_fdbuf_t *buf);
 static int		mxml_file_getc(void *p, int *encoding);
 static int		mxml_file_putc(int ch, void *p);
 static int		mxml_get_entity(mxml_node_t *parent, void *p,
 			                int *encoding,
-					int (*getc_cb)(void *, int *));
+					_mxml_getc_cb_t getc_cb);
+static inline int	mxml_isspace(int ch)
+			{
+			  return (ch == ' ' || ch == '\t' || ch == '\r' ||
+			          ch == '\n');
+			}
 static mxml_node_t	*mxml_load_data(mxml_node_t *top, void *p,
-			                mxml_type_t (*cb)(mxml_node_t *),
-			                int (*getc_cb)(void *, int *));
+			                mxml_load_cb_t cb,
+			                _mxml_getc_cb_t getc_cb,
+                                        mxml_sax_cb_t sax_cb, void *sax_data);
 static int		mxml_parse_element(mxml_node_t *node, void *p,
 			                   int *encoding,
-					   int (*getc_cb)(void *, int *));
+					   _mxml_getc_cb_t getc_cb);
 static int		mxml_string_getc(void *p, int *encoding);
 static int		mxml_string_putc(int ch, void *p);
 static int		mxml_write_name(const char *s, void *p,
-					int (*putc_cb)(int, void *));
+					_mxml_putc_cb_t putc_cb);
 static int		mxml_write_node(mxml_node_t *node, void *p,
-			                const char *(*cb)(mxml_node_t *, int),
-					int col,
-					int (*putc_cb)(int, void *));
+			                mxml_save_cb_t cb, int col,
+					_mxml_putc_cb_t putc_cb,
+					_mxml_global_t *global);
 static int		mxml_write_string(const char *s, void *p,
-					  int (*putc_cb)(int, void *));
+					  _mxml_putc_cb_t putc_cb);
 static int		mxml_write_ws(mxml_node_t *node, void *p, 
-			              const char *(*cb)(mxml_node_t *, int), int ws,
-				      int col, int (*putc_cb)(int, void *));
+			              mxml_save_cb_t cb, int ws,
+				      int col, _mxml_putc_cb_t putc_cb);
 
 
 /*
@@ -153,12 +153,11 @@ static int		mxml_write_ws(mxml_node_t *node, void *p,
  */
 
 mxml_node_t *				/* O - First node or NULL if the file could not be read. */
-mxmlLoadFd(mxml_node_t *top,		/* I - Top node */
-           int         fd,		/* I - File descriptor to read from */
-           mxml_type_t (*cb)(mxml_node_t *node))
-					/* I - Callback function or MXML_NO_CALLBACK */
+mxmlLoadFd(mxml_node_t    *top,		/* I - Top node */
+           int            fd,		/* I - File descriptor to read from */
+           mxml_load_cb_t cb)		/* I - Callback function or MXML_NO_CALLBACK */
 {
-  mxml_fdbuf_t	buf;			/* File descriptor buffer */
+  _mxml_fdbuf_t	buf;			/* File descriptor buffer */
 
 
  /*
@@ -173,7 +172,7 @@ mxmlLoadFd(mxml_node_t *top,		/* I - Top node */
   * Read the XML data...
   */
 
-  return (mxml_load_data(top, &buf, cb, mxml_fd_getc));
+  return (mxml_load_data(top, &buf, cb, mxml_fd_getc, MXML_NO_CALLBACK, NULL));
 }
 
 
@@ -193,16 +192,15 @@ mxmlLoadFd(mxml_node_t *top,		/* I - Top node */
  */
 
 mxml_node_t *				/* O - First node or NULL if the file could not be read. */
-mxmlLoadFile(mxml_node_t *top,		/* I - Top node */
-             FILE        *fp,		/* I - File to read from */
-             mxml_type_t (*cb)(mxml_node_t *node))
-					/* I - Callback function or MXML_NO_CALLBACK */
+mxmlLoadFile(mxml_node_t    *top,	/* I - Top node */
+             FILE           *fp,	/* I - File to read from */
+             mxml_load_cb_t cb)		/* I - Callback function or MXML_NO_CALLBACK */
 {
  /*
   * Read the XML data...
   */
 
-  return (mxml_load_data(top, fp, cb, mxml_file_getc));
+  return (mxml_load_data(top, fp, cb, mxml_file_getc, MXML_NO_CALLBACK, NULL));
 }
 
 
@@ -222,16 +220,16 @@ mxmlLoadFile(mxml_node_t *top,		/* I - Top node */
  */
 
 mxml_node_t *				/* O - First node or NULL if the string has errors. */
-mxmlLoadString(mxml_node_t *top,	/* I - Top node */
-               const char  *s,		/* I - String to load */
-               mxml_type_t (*cb)(mxml_node_t *node))
-					/* I - Callback function or MXML_NO_CALLBACK */
+mxmlLoadString(mxml_node_t    *top,	/* I - Top node */
+               const char     *s,	/* I - String to load */
+               mxml_load_cb_t cb)	/* I - Callback function or MXML_NO_CALLBACK */
 {
  /*
   * Read the XML data...
   */
 
-  return (mxml_load_data(top, &s, cb, mxml_string_getc));
+  return (mxml_load_data(top, (void *)&s, cb, mxml_string_getc, MXML_NO_CALLBACK,
+                         NULL));
 }
 
 
@@ -252,9 +250,9 @@ mxmlLoadString(mxml_node_t *top,	/* I - Top node */
  */
 
 char *					/* O - Allocated string or NULL */
-mxmlSaveAllocString(mxml_node_t *node,	/* I - Node to write */
-                    const char  *(*cb)(mxml_node_t *node, int ws))
-					/* I - Whitespace callback or MXML_NO_CALLBACK */
+mxmlSaveAllocString(
+    mxml_node_t    *node,		/* I - Node to write */
+    mxml_save_cb_t cb)			/* I - Whitespace callback or MXML_NO_CALLBACK */
 {
   int	bytes;				/* Required bytes */
   char	buffer[8192];			/* Temporary buffer */
@@ -309,13 +307,14 @@ mxmlSaveAllocString(mxml_node_t *node,	/* I - Node to write */
  */
 
 int					/* O - 0 on success, -1 on error. */
-mxmlSaveFd(mxml_node_t *node,		/* I - Node to write */
-           int         fd,		/* I - File descriptor to write to */
-	   const char  *(*cb)(mxml_node_t *node, int ws))
-					/* I - Whitespace callback or MXML_NO_CALLBACK */
+mxmlSaveFd(mxml_node_t    *node,	/* I - Node to write */
+           int            fd,		/* I - File descriptor to write to */
+	   mxml_save_cb_t cb)		/* I - Whitespace callback or MXML_NO_CALLBACK */
 {
   int		col;			/* Final column */
-  mxml_fdbuf_t	buf;			/* File descriptor buffer */
+  _mxml_fdbuf_t	buf;			/* File descriptor buffer */
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
 
 
  /*
@@ -330,7 +329,7 @@ mxmlSaveFd(mxml_node_t *node,		/* I - Node to write */
   * Write the node...
   */
 
-  if ((col = mxml_write_node(node, &buf, cb, 0, mxml_fd_putc)) < 0)
+  if ((col = mxml_write_node(node, &buf, cb, 0, mxml_fd_putc, global)) < 0)
     return (-1);
 
   if (col > 0)
@@ -356,19 +355,20 @@ mxmlSaveFd(mxml_node_t *node,		/* I - Node to write */
  */
 
 int					/* O - 0 on success, -1 on error. */
-mxmlSaveFile(mxml_node_t *node,		/* I - Node to write */
-             FILE        *fp,		/* I - File to write to */
-	     const char  *(*cb)(mxml_node_t *node, int ws))
-					/* I - Whitespace callback or MXML_NO_CALLBACK */
+mxmlSaveFile(mxml_node_t    *node,	/* I - Node to write */
+             FILE           *fp,	/* I - File to write to */
+	     mxml_save_cb_t cb)		/* I - Whitespace callback or MXML_NO_CALLBACK */
 {
   int	col;				/* Final column */
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
 
 
  /*
   * Write the node...
   */
 
-  if ((col = mxml_write_node(node, fp, cb, 0, mxml_file_putc)) < 0)
+  if ((col = mxml_write_node(node, fp, cb, 0, mxml_file_putc, global)) < 0)
     return (-1);
 
   if (col > 0)
@@ -398,14 +398,15 @@ mxmlSaveFile(mxml_node_t *node,		/* I - Node to write */
  */
 
 int					/* O - Size of string */
-mxmlSaveString(mxml_node_t *node,	/* I - Node to write */
-               char        *buffer,	/* I - String buffer */
-               int         bufsize,	/* I - Size of string buffer */
-               const char  *(*cb)(mxml_node_t *node, int ws))
-					/* I - Whitespace callback or MXML_NO_CALLBACK */
+mxmlSaveString(mxml_node_t    *node,	/* I - Node to write */
+               char           *buffer,	/* I - String buffer */
+               int            bufsize,	/* I - Size of string buffer */
+               mxml_save_cb_t cb)	/* I - Whitespace callback or MXML_NO_CALLBACK */
 {
   int	col;				/* Final column */
   char	*ptr[2];			/* Pointers for putc_cb */
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
 
 
  /*
@@ -415,7 +416,7 @@ mxmlSaveString(mxml_node_t *node,	/* I - Node to write */
   ptr[0] = buffer;
   ptr[1] = buffer + bufsize;
 
-  if ((col = mxml_write_node(node, ptr, cb, 0, mxml_string_putc)) < 0)
+  if ((col = mxml_write_node(node, ptr, cb, 0, mxml_string_putc, global)) < 0)
     return (-1);
 
   if (col > 0)
@@ -438,6 +439,130 @@ mxmlSaveString(mxml_node_t *node,	/* I - Node to write */
 }
 
 
+/*
+ * 'mxmlSAXLoadFd()' - Load a file descriptor into an XML node tree
+ *                     using a SAX callback.
+ *
+ * The nodes in the specified file are added to the specified top node.
+ * If no top node is provided, the XML file MUST be well-formed with a
+ * single parent node like <?xml> for the entire file. The callback
+ * function returns the value type that should be used for child nodes.
+ * If MXML_NO_CALLBACK is specified then all child nodes will be either
+ * MXML_ELEMENT or MXML_TEXT nodes.
+ *
+ * The constants MXML_INTEGER_CALLBACK, MXML_OPAQUE_CALLBACK,
+ * MXML_REAL_CALLBACK, and MXML_TEXT_CALLBACK are defined for loading
+ * child nodes of the specified type.
+ *
+ * The SAX callback must call mxmlRetain() for any nodes that need to
+ * be kept for later use. Otherwise, nodes are deleted when the parent
+ * node is closed or after each data, comment, CDATA, or directive node.
+ *
+ * @since Mini-XML 2.3@
+ */
+
+mxml_node_t *				/* O - First node or NULL if the file could not be read. */
+mxmlSAXLoadFd(mxml_node_t    *top,	/* I - Top node */
+              int            fd,	/* I - File descriptor to read from */
+              mxml_load_cb_t cb,	/* I - Callback function or MXML_NO_CALLBACK */
+              mxml_sax_cb_t  sax_cb,	/* I - SAX callback or MXML_NO_CALLBACK */
+              void           *sax_data)	/* I - SAX user data */
+{
+  _mxml_fdbuf_t	buf;			/* File descriptor buffer */
+
+
+ /*
+  * Initialize the file descriptor buffer...
+  */
+
+  buf.fd      = fd;
+  buf.current = buf.buffer;
+  buf.end     = buf.buffer;
+
+ /*
+  * Read the XML data...
+  */
+
+  return (mxml_load_data(top, &buf, cb, mxml_fd_getc, sax_cb, sax_data));
+}
+
+
+/*
+ * 'mxmlSAXLoadFile()' - Load a file into an XML node tree
+ *                       using a SAX callback.
+ *
+ * The nodes in the specified file are added to the specified top node.
+ * If no top node is provided, the XML file MUST be well-formed with a
+ * single parent node like <?xml> for the entire file. The callback
+ * function returns the value type that should be used for child nodes.
+ * If MXML_NO_CALLBACK is specified then all child nodes will be either
+ * MXML_ELEMENT or MXML_TEXT nodes.
+ *
+ * The constants MXML_INTEGER_CALLBACK, MXML_OPAQUE_CALLBACK,
+ * MXML_REAL_CALLBACK, and MXML_TEXT_CALLBACK are defined for loading
+ * child nodes of the specified type.
+ *
+ * The SAX callback must call mxmlRetain() for any nodes that need to
+ * be kept for later use. Otherwise, nodes are deleted when the parent
+ * node is closed or after each data, comment, CDATA, or directive node.
+ *
+ * @since Mini-XML 2.3@
+ */
+
+mxml_node_t *				/* O - First node or NULL if the file could not be read. */
+mxmlSAXLoadFile(
+    mxml_node_t    *top,		/* I - Top node */
+    FILE           *fp,			/* I - File to read from */
+    mxml_load_cb_t cb,			/* I - Callback function or MXML_NO_CALLBACK */
+    mxml_sax_cb_t  sax_cb,		/* I - SAX callback or MXML_NO_CALLBACK */
+    void           *sax_data)		/* I - SAX user data */
+{
+ /*
+  * Read the XML data...
+  */
+
+  return (mxml_load_data(top, fp, cb, mxml_file_getc, sax_cb, sax_data));
+}
+
+
+/*
+ * 'mxmlSAXLoadString()' - Load a string into an XML node tree
+ *                         using a SAX callback.
+ *
+ * The nodes in the specified string are added to the specified top node.
+ * If no top node is provided, the XML string MUST be well-formed with a
+ * single parent node like <?xml> for the entire string. The callback
+ * function returns the value type that should be used for child nodes.
+ * If MXML_NO_CALLBACK is specified then all child nodes will be either
+ * MXML_ELEMENT or MXML_TEXT nodes.
+ *
+ * The constants MXML_INTEGER_CALLBACK, MXML_OPAQUE_CALLBACK,
+ * MXML_REAL_CALLBACK, and MXML_TEXT_CALLBACK are defined for loading
+ * child nodes of the specified type.
+ *
+ * The SAX callback must call mxmlRetain() for any nodes that need to
+ * be kept for later use. Otherwise, nodes are deleted when the parent
+ * node is closed or after each data, comment, CDATA, or directive node.
+ *
+ * @since Mini-XML 2.3@
+ */
+
+mxml_node_t *				/* O - First node or NULL if the string has errors. */
+mxmlSAXLoadString(
+    mxml_node_t    *top,		/* I - Top node */
+    const char     *s,			/* I - String to load */
+    mxml_load_cb_t cb,			/* I - Callback function or MXML_NO_CALLBACK */
+    mxml_sax_cb_t  sax_cb,		/* I - SAX callback or MXML_NO_CALLBACK */
+    void           *sax_data)		/* I - SAX user data */
+{
+ /*
+  * Read the XML data...
+  */
+
+  return (mxml_load_data(top, (void *)&s, cb, mxml_string_getc, sax_cb, sax_data));
+}
+
+
 /*
  * 'mxmlSetCustomHandlers()' - Set the handling functions for custom data.
  *
@@ -450,13 +575,16 @@ mxmlSaveString(mxml_node_t *node,	/* I - Node to write */
  */
 
 void
-mxmlSetCustomHandlers(mxml_custom_load_cb_t load,
-					/* I - Load function */
-                      mxml_custom_save_cb_t save)
-					/* I - Save function */
+mxmlSetCustomHandlers(
+    mxml_custom_load_cb_t load,		/* I - Load function */
+    mxml_custom_save_cb_t save)		/* I - Save function */
 {
-  mxml_custom_load_cb = load;
-  mxml_custom_save_cb = save;
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
+
+
+  global->custom_load_cb = load;
+  global->custom_save_cb = save;
 }
 
 
@@ -465,10 +593,35 @@ mxmlSetCustomHandlers(mxml_custom_load_cb_t load,
  */
 
 void
-mxmlSetErrorCallback(void (*cb)(const char *))
-					/* I - Error callback function */
+mxmlSetErrorCallback(mxml_error_cb_t cb)/* I - Error callback function */
 {
-  mxml_error_cb = cb;
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
+
+
+  global->error_cb = cb;
+}
+
+
+/*
+ * 'mxmlSetWrapMargin()' - Set the the wrap margin when saving XML data.
+ *
+ * Wrapping is disabled when "column" is <= 0.
+ *
+ * @since Mini-XML 2.3@
+ */
+
+void
+mxmlSetWrapMargin(int column)		/* I - Column for wrapping */
+{
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
+
+
+  if (column <= 0)
+    global->wrap = 2147483647;
+  else
+    global->wrap = column;
 }
 
 
@@ -560,7 +713,7 @@ static int				/* O  - Character or EOF */
 mxml_fd_getc(void *p,			/* I  - File descriptor buffer */
              int  *encoding)		/* IO - Encoding */
 {
-  mxml_fdbuf_t	*buf;			/* File descriptor buffer */
+  _mxml_fdbuf_t	*buf;			/* File descriptor buffer */
   int		ch,			/* Current character */
 		temp;			/* Temporary character */
 
@@ -569,7 +722,7 @@ mxml_fd_getc(void *p,			/* I  - File descriptor buffer */
   * Grab the next character in the buffer...
   */
 
-  buf = (mxml_fdbuf_t *)p;
+  buf = (_mxml_fdbuf_t *)p;
 
   if (buf->current >= buf->end)
     if (mxml_fd_read(buf) < 0)
@@ -847,7 +1000,7 @@ static int				/* O - 0 on success, -1 on error */
 mxml_fd_putc(int  ch,			/* I - Character */
              void *p)			/* I - File descriptor buffer */
 {
-  mxml_fdbuf_t	*buf;			/* File descriptor buffer */
+  _mxml_fdbuf_t	*buf;			/* File descriptor buffer */
 
 
  /*
@@ -855,7 +1008,7 @@ mxml_fd_putc(int  ch,			/* I - Character */
   * 4 characters at the end so that we can avoid a lot of extra tests...
   */
 
-  buf = (mxml_fdbuf_t *)p;
+  buf = (_mxml_fdbuf_t *)p;
 
   if (buf->current >= buf->end)
     if (mxml_fd_write(buf) < 0)
@@ -913,7 +1066,7 @@ mxml_fd_putc(int  ch,			/* I - Character */
  */
 
 static int				/* O - 0 on success, -1 on error */
-mxml_fd_read(mxml_fdbuf_t *buf)		/* I - File descriptor buffer */
+mxml_fd_read(_mxml_fdbuf_t *buf)		/* I - File descriptor buffer */
 {
   int	bytes;				/* Bytes read... */
 
@@ -930,7 +1083,11 @@ mxml_fd_read(mxml_fdbuf_t *buf)		/* I - File descriptor buffer */
   */
 
   while ((bytes = read(buf->fd, buf->buffer, sizeof(buf->buffer))) < 0)
+#ifdef EINTR
     if (errno != EAGAIN && errno != EINTR)
+#else
+    if (errno != EAGAIN)
+#endif /* EINTR */
       return (-1);
 
   if (bytes == 0)
@@ -952,7 +1109,7 @@ mxml_fd_read(mxml_fdbuf_t *buf)		/* I - File descriptor buffer */
  */
 
 static int				/* O - 0 on success, -1 on error */
-mxml_fd_write(mxml_fdbuf_t *buf)	/* I - File descriptor buffer */
+mxml_fd_write(_mxml_fdbuf_t *buf)	/* I - File descriptor buffer */
 {
   int		bytes;			/* Bytes written */
   unsigned char	*ptr;			/* Pointer into buffer */
@@ -1310,12 +1467,13 @@ mxml_get_entity(mxml_node_t *parent,	/* I  - Parent node */
  */
 
 static mxml_node_t *			/* O - First node or NULL if the file could not be read. */
-mxml_load_data(mxml_node_t *top,	/* I - Top node */
-               void        *p,		/* I - Pointer to data */
-               mxml_type_t (*cb)(mxml_node_t *),
-					/* I - Callback function or MXML_NO_CALLBACK */
-               int         (*getc_cb)(void *, int *))
-					/* I - Read function */
+mxml_load_data(
+    mxml_node_t     *top,		/* I - Top node */
+    void            *p,			/* I - Pointer to data */
+    mxml_load_cb_t  cb,			/* I - Callback function or MXML_NO_CALLBACK */
+    _mxml_getc_cb_t getc_cb,		/* I - Read function */
+    mxml_sax_cb_t   sax_cb,		/* I - SAX callback or MXML_NO_CALLBACK */
+    void            *sax_data)		/* I - SAX user data */
 {
   mxml_node_t	*node,			/* Current node */
 		*first,			/* First node added */
@@ -1327,6 +1485,8 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
   int		bufsize;		/* Size of buffer */
   mxml_type_t	type;			/* Current node type */
   int		encoding;		/* Character encoding */
+  _mxml_global_t *global = _mxml_global();
+					/* Global data */
   static const char * const types[] =	/* Type strings... */
 		{
 		  "MXML_ELEMENT",	/* XML element with attributes */
@@ -1363,7 +1523,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
   while ((ch = (*getc_cb)(p, &encoding)) != EOF)
   {
     if ((ch == '<' ||
-         (isspace(ch) && type != MXML_OPAQUE && type != MXML_CUSTOM)) &&
+         (mxml_isspace(ch) && type != MXML_OPAQUE && type != MXML_CUSTOM)) &&
         bufptr > buffer)
     {
      /*
@@ -1391,7 +1551,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	    break;
 
 	case MXML_CUSTOM :
-	    if (mxml_custom_load_cb)
+	    if (global->custom_load_cb)
 	    {
 	     /*
 	      * Use the callback to fill in the custom data...
@@ -1399,7 +1559,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
               node = mxmlNewCustom(parent, NULL, NULL);
 
-	      if ((*mxml_custom_load_cb)(node, buffer))
+	      if ((*global->custom_load_cb)(node, buffer))
 	      {
 	        mxml_error("Bad custom value '%s' in parent <%s>!",
 		           buffer, parent ? parent->value.element.name : "null");
@@ -1409,7 +1569,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	      break;
 	    }
 
-        default : /* Should never happen... */
+        default : /* Ignore... */
 	    node = NULL;
 	    break;
       }	  
@@ -1427,9 +1587,9 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
       }
 
       bufptr     = buffer;
-      whitespace = isspace(ch) && type == MXML_TEXT;
+      whitespace = mxml_isspace(ch) && type == MXML_TEXT;
 
-      if (!node)
+      if (!node && type != MXML_IGNORE)
       {
        /*
 	* Print error and return...
@@ -1440,10 +1600,18 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	goto error;
       }
 
-      if (!first)
+      if (sax_cb)
+      {
+        (*sax_cb)(node, MXML_SAX_DATA, sax_data);
+
+        if (!mxmlRelease(node))
+          node = NULL;
+      }
+
+      if (!first && node)
         first = node;
     }
-    else if (isspace(ch) && type == MXML_TEXT)
+    else if (mxml_isspace(ch) && type == MXML_TEXT)
       whitespace = 1;
 
    /*
@@ -1453,7 +1621,19 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
     if (ch == '<' && whitespace && type == MXML_TEXT)
     {
-      mxmlNewText(parent, whitespace, "");
+      node = mxmlNewText(parent, whitespace, "");
+
+      if (sax_cb)
+      {
+        (*sax_cb)(node, MXML_SAX_DATA, sax_data);
+
+        if (!mxmlRelease(node))
+          node = NULL;
+      }
+
+      if (!first && node)
+        first = node;
+
       whitespace = 0;
     }
 
@@ -1466,8 +1646,13 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
       bufptr = buffer;
 
       while ((ch = (*getc_cb)(p, &encoding)) != EOF)
-        if (isspace(ch) || ch == '>' || (ch == '/' && bufptr > buffer))
+        if (mxml_isspace(ch) || ch == '>' || (ch == '/' && bufptr > buffer))
 	  break;
+	else if (ch == '<')
+	{
+	  mxml_error("Bare < in element!");
+	  goto error;
+	}
 	else if (ch == '&')
 	{
 	  if ((ch = mxml_get_entity(parent, p, &encoding, getc_cb)) == EOF)
@@ -1521,7 +1706,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
 	*bufptr = '\0';
 
-	if (!mxmlNewElement(parent, buffer))
+	if ((node = mxmlNewElement(parent, buffer)) == NULL)
 	{
 	 /*
 	  * Just print error for now...
@@ -1531,6 +1716,17 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	             parent ? parent->value.element.name : "null");
 	  break;
 	}
+
+        if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_COMMENT, sax_data);
+
+          if (!mxmlRelease(node))
+            node = NULL;
+        }
+
+	if (node && !first)
+	  first = node;
       }
       else if (!strcmp(buffer, "![CDATA["))
       {
@@ -1567,7 +1763,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
 	*bufptr = '\0';
 
-	if (!mxmlNewElement(parent, buffer))
+	if ((node = mxmlNewElement(parent, buffer)) == NULL)
 	{
 	 /*
 	  * Print error and return...
@@ -1577,6 +1773,17 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	             parent ? parent->value.element.name : "null");
 	  goto error;
 	}
+
+        if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_CDATA, sax_data);
+
+          if (!mxmlRelease(node))
+            node = NULL;
+        }
+
+	if (node && !first)
+	  first = node;
       }
       else if (buffer[0] == '?')
       {
@@ -1606,14 +1813,13 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  goto error;
 	}
 
-
        /*
         * Otherwise add this as an element under the current parent...
 	*/
 
 	*bufptr = '\0';
 
-	if (!(parent = mxmlNewElement(parent, buffer)))
+	if ((node = mxmlNewElement(parent, buffer)) == NULL)
 	{
 	 /*
 	  * Print error and return...
@@ -1624,8 +1830,24 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  goto error;
 	}
 
-	if (cb)
-	  type = (*cb)(parent);
+        if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_DIRECTIVE, sax_data);
+
+          if (!mxmlRelease(node))
+            node = NULL;
+        }
+
+        if (node)
+	{
+	  if (!first)
+	    first = node;
+
+	  parent = node;
+
+	  if (cb)
+	    type = (*cb)(parent);
+        }
       }
       else if (buffer[0] == '!')
       {
@@ -1669,8 +1891,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
 	*bufptr = '\0';
 
-	node = mxmlNewElement(parent, buffer);
-	if (!node)
+	if ((node = mxmlNewElement(parent, buffer)) == NULL)
 	{
 	 /*
 	  * Print error and return...
@@ -1681,14 +1902,16 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  goto error;
 	}
 
-       /*
-	* Descend into this node, setting the value type as needed...
-	*/
+        if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_DIRECTIVE, sax_data);
 
-	parent = node;
+          if (!mxmlRelease(node))
+            node = NULL;
+        }
 
-	if (cb)
-	  type = (*cb)(parent);
+        if (node && !first)
+          first = node;
       }
       else if (buffer[0] == '/')
       {
@@ -1703,7 +1926,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  */
 
 	  mxml_error("Mismatched close tag <%s> under parent <%s>!",
-	             buffer, parent->value.element.name);
+	             buffer, parent ? parent->value.element.name : "(null)");
           goto error;
 	}
 
@@ -1714,12 +1937,20 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
         while (ch != '>' && ch != EOF)
 	  ch = (*getc_cb)(p, &encoding);
 
+        node   = parent;
+        parent = parent->parent;
+
+        if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_ELEMENT_CLOSE, sax_data);
+
+          mxmlRelease(node);
+        }
+
        /*
 	* Ascend into the parent and set the value type as needed...
 	*/
 
-	parent = parent->parent;
-
 	if (cb && parent)
 	  type = (*cb)(parent);
       }
@@ -1729,9 +1960,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
         * Handle open tag...
 	*/
 
-        node = mxmlNewElement(parent, buffer);
-
-	if (!node)
+        if ((node = mxmlNewElement(parent, buffer)) == NULL)
 	{
 	 /*
 	  * Just print error for now...
@@ -1742,20 +1971,30 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  goto error;
 	}
 
-        if (isspace(ch))
-          ch = mxml_parse_element(node, p, &encoding, getc_cb);
+        if (mxml_isspace(ch))
+        {
+	  if ((ch = mxml_parse_element(node, p, &encoding, getc_cb)) == EOF)
+	    goto error;
+        }
         else if (ch == '/')
 	{
 	  if ((ch = (*getc_cb)(p, &encoding)) != '>')
 	  {
 	    mxml_error("Expected > but got '%c' instead for element <%s/>!",
 	               ch, buffer);
+            mxmlDelete(node);
             goto error;
 	  }
 
 	  ch = '/';
 	}
 
+        if (sax_cb)
+          (*sax_cb)(node, MXML_SAX_ELEMENT_OPEN, sax_data);
+
+        if (!first)
+	  first = node;
+
 	if (ch == EOF)
 	  break;
 
@@ -1770,6 +2009,13 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 	  if (cb && parent)
 	    type = (*cb)(parent);
 	}
+        else if (sax_cb)
+        {
+          (*sax_cb)(node, MXML_SAX_ELEMENT_CLOSE, sax_data);
+
+          if (!mxmlRelease(node) && first == node)
+            first = NULL;
+        }
       }
 
       bufptr  = buffer;
@@ -1786,7 +2032,7 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
       if (mxml_add_char(ch, &bufptr, &buffer, &bufsize))
 	goto error;
     }
-    else if (type == MXML_OPAQUE || type == MXML_CUSTOM || !isspace(ch))
+    else if (type == MXML_OPAQUE || type == MXML_CUSTOM || !mxml_isspace(ch))
     {
      /*
       * Add character to current buffer...
@@ -1809,11 +2055,27 @@ mxml_load_data(mxml_node_t *top,	/* I - Top node */
 
   if (parent)
   {
+    node = parent;
+
     while (parent->parent != top && parent->parent)
       parent = parent->parent;
+
+    if (node != parent)
+    {
+      mxml_error("Missing close tag </%s> under parent <%s>!",
+	         node->value.element.name,
+		 node->parent ? node->parent->value.element.name : "(null)");
+
+      mxmlDelete(first);
+
+      return (NULL);
+    }
   }
 
-  return (parent);
+  if (parent)
+    return (parent);
+  else
+    return (first);
 
  /*
   * Common error return...
@@ -1834,12 +2096,11 @@ error:
  */
 
 static int				/* O  - Terminating character */
-mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
-                   void        *p,	/* I  - Data to read from */
-		   int         *encoding,
-					/* IO - Encoding */
-                   int         (*getc_cb)(void *, int *))
-					/* I  - Data callback */
+mxml_parse_element(
+    mxml_node_t     *node,		/* I  - Element node */
+    void            *p,			/* I  - Data to read from */
+    int             *encoding,		/* IO - Encoding */
+    _mxml_getc_cb_t getc_cb)		/* I  - Data callback */
 {
   int	ch,				/* Current character in file */
 	quote;				/* Quoting character */
@@ -1887,7 +2148,7 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
     * Skip leading whitespace...
     */
 
-    if (isspace(ch))
+    if (mxml_isspace(ch))
       continue;
 
    /*
@@ -1906,11 +2167,16 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
       {
         mxml_error("Expected '>' after '%c' for element %s, but got '%c'!",
 	           ch, node->value.element.name, quote);
-        ch = EOF;
+        goto error;
       }
 
       break;
     }
+    else if (ch == '<')
+    {
+      mxml_error("Bare < in element %s!", node->value.element.name);
+      goto error;
+    }
     else if (ch == '>')
       break;
 
@@ -1949,7 +2215,8 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
       */
 
       while ((ch = (*getc_cb)(p, encoding)) != EOF)
-	if (isspace(ch) || ch == '=' || ch == '/' || ch == '>' || ch == '?')
+	if (mxml_isspace(ch) || ch == '=' || ch == '/' || ch == '>' ||
+	    ch == '?')
           break;
 	else
 	{
@@ -1977,7 +2244,7 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
       {
         mxml_error("Missing value for attribute '%s' in element %s!",
 	           name, node->value.element.name);
-        return (EOF);
+        goto error;
       }
 
       if (ch == '\'' || ch == '\"')
@@ -2014,7 +2281,7 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
 	ptr      = value + 1;
 
 	while ((ch = (*getc_cb)(p, encoding)) != EOF)
-	  if (isspace(ch) || ch == '=' || ch == '/' || ch == '>')
+	  if (mxml_isspace(ch) || ch == '=' || ch == '/' || ch == '>')
             break;
 	  else
 	  {
@@ -2037,11 +2304,9 @@ mxml_parse_element(mxml_node_t *node,	/* I  - Element node */
     }
     else
     {
-     /*
-      * Set the attribute with a NULL value...
-      */
-
-      mxmlElementSetAttr(node, name, NULL);
+      mxml_error("Missing value for attribute '%s' in element %s!",
+	         name, node->value.element.name);
+      goto error;
     }
 
    /*
@@ -2481,12 +2746,12 @@ mxml_write_name(const char *s,		/* I - Name to write */
  */
 
 static int				/* O - Column or -1 on error */
-mxml_write_node(mxml_node_t *node,	/* I - Node to write */
-                void        *p,		/* I - File to write to */
-	        const char  *(*cb)(mxml_node_t *, int),
-					/* I - Whitespace callback */
-		int         col,	/* I - Current column */
-		int         (*putc_cb)(int, void *))
+mxml_write_node(mxml_node_t     *node,	/* I - Node to write */
+                void            *p,	/* I - File to write to */
+	        mxml_save_cb_t  cb,	/* I - Whitespace callback */
+		int             col,	/* I - Current column */
+		_mxml_putc_cb_t putc_cb,/* I - Output callback */
+		_mxml_global_t  *global)/* I - Global data */
 {
   int		i,			/* Looping var */
 		width;			/* Width of attr + value */
@@ -2528,7 +2793,7 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
 	    */
 
             if (!strncmp(node->value.element.name, "?xml", 4))
-              col = MXML_WRAP;
+              col = global->wrap;
 	  }
 	  else if (mxml_write_name(node->value.element.name, p, putc_cb) < 0)
 	    return (-1);
@@ -2544,7 +2809,7 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
 	    if (attr->value)
 	      width += strlen(attr->value) + 3;
 
-	    if ((col + width) > MXML_WRAP)
+	    if ((col + width) > global->wrap)
 	    {
 	      if ((*putc_cb)('\n', p) < 0)
 	        return (-1);
@@ -2590,7 +2855,8 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
 
             col = mxml_write_ws(node, p, cb, MXML_WS_AFTER_OPEN, col, putc_cb);
 
-	    if ((col = mxml_write_node(node->child, p, cb, col, putc_cb)) < 0)
+	    if ((col = mxml_write_node(node->child, p, cb, col, putc_cb,
+	                               global)) < 0)
 	      return (-1);
 
            /*
@@ -2648,7 +2914,7 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
       case MXML_INTEGER :
 	  if (node->prev)
 	  {
-	    if (col > MXML_WRAP)
+	    if (col > global->wrap)
 	    {
 	      if ((*putc_cb)('\n', p) < 0)
 	        return (-1);
@@ -2678,7 +2944,7 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
       case MXML_REAL :
 	  if (node->prev)
 	  {
-	    if (col > MXML_WRAP)
+	    if (col > global->wrap)
 	    {
 	      if ((*putc_cb)('\n', p) < 0)
 	        return (-1);
@@ -2701,7 +2967,7 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
       case MXML_TEXT :
 	  if (node->value.text.whitespace && col > 0)
 	  {
-	    if (col > MXML_WRAP)
+	    if (col > global->wrap)
 	    {
 	      if ((*putc_cb)('\n', p) < 0)
 	        return (-1);
@@ -2721,13 +2987,13 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
           break;
 
       case MXML_CUSTOM :
-          if (mxml_custom_save_cb)
+          if (global->custom_save_cb)
 	  {
 	    char	*data;		/* Custom data string */
 	    const char	*newline;	/* Last newline in string */
 
 
-            if ((data = (*mxml_custom_save_cb)(node)) == NULL)
+            if ((data = (*global->custom_save_cb)(node)) == NULL)
 	      return (-1);
 
             if (mxml_write_string(data, p, putc_cb) < 0)
@@ -2762,10 +3028,10 @@ mxml_write_node(mxml_node_t *node,	/* I - Node to write */
  */
 
 static int				/* O - 0 on success, -1 on failure */
-mxml_write_string(const char *s,	/* I - String to write */
-                  void       *p,	/* I - Write pointer */
-		  int        (*putc_cb)(int, void *))
-					/* I - Write callback */
+mxml_write_string(
+    const char      *s,			/* I - String to write */
+    void            *p,			/* I - Write pointer */
+    _mxml_putc_cb_t putc_cb)		/* I - Write callback */
 {
   const char	*name;			/* Entity name, if any */
 
@@ -2802,14 +3068,12 @@ mxml_write_string(const char *s,	/* I - String to write */
  */
 
 static int				/* O - New column */
-mxml_write_ws(mxml_node_t *node,	/* I - Current node */
-              void        *p,		/* I - Write pointer */
-              const char  *(*cb)(mxml_node_t *, int),
-					/* I - Callback function */
-	      int         ws,		/* I - Where value */
-	      int         col,		/* I - Current column */
-              int         (*putc_cb)(int, void *))
-					/* I - Write callback */
+mxml_write_ws(mxml_node_t     *node,	/* I - Current node */
+              void            *p,	/* I - Write pointer */
+              mxml_save_cb_t  cb,	/* I - Callback function */
+	      int             ws,	/* I - Where value */
+	      int             col,	/* I - Current column */
+              _mxml_putc_cb_t putc_cb)	/* I - Write callback */
 {
   const char	*s;			/* Whitespace string */
 
@@ -2839,5 +3103,5 @@ mxml_write_ws(mxml_node_t *node,	/* I - Current node */
 
 
 /*
- * End of "$Id: mxml-file.c 210 2005-05-19 14:56:05Z mike $".
+ * End of "$Id: mxml-file.c 312 2007-10-03 06:25:07Z mike $".
  */
