@@ -1,5 +1,7 @@
 /*
  * 'foo_void_function()' - Do foo with bar.
+ *
+ * @deprecated@
  */
 
 void
@@ -20,6 +22,8 @@ foo_void_function(int        one,	/* I - Integer */
 
 /*
  * 'foo_float_function()' - Do foo with bar.
+ *
+ * @since 1.2@
  */
 
 float					/* O - Real number */
@@ -74,3 +78,27 @@ foo_default_int(int one,		/* I - Integer */
 
   return (2);
 }
+
+
+/*
+ * 'foo_void_func()' - Function taking no arguments.
+ */
+
+void
+foo_void_func(void)
+{
+  puts("foo_void_func()");
+}
+
+
+/*
+ * 'foo_private_func()' - Private function.
+ *
+ * @private@
+ */
+
+void
+foo_private_func(void)
+{
+  puts("foo_private_func()");
+}
